from pydantic import Field

from schorle.elements.base.element import Element
from schorle.elements.html import Div
from schorle.elements.img import Img
from schorle.elements.tags import HTMLTag
from schorle.reactives.classes import Classes


class Figure(Element):
    tag: HTMLTag = HTMLTag.FIGURE


class FigureWithImg(Figure):
    img: Img


class CardTitle(Element):
    tag: HTMLTag = HTMLTag.H2
    _base_classes: Classes = Classes("card-title")


class CardActions(Div):
    _base_classes: Classes = Classes("card-actions")


class CardBody(Div):
    _base_classes: Classes = Classes("card-body")
    title: CardTitle = Field(default_factory=CardTitle)
    body: Div = Field(default_factory=Div)
    actions: CardActions | None = None


class Card(Div):
    _base_classes: Classes = Classes("card")
    figure: Figure | None = None
    body: CardBody = Field(default_factory=CardBody)



import asyncio
from asyncio import Protocol
from functools import wraps
from inspect import ismethod

from pydantic import BaseModel


class Effector:
    def __init__(self, bounded_method):
        self.bounded_method = bounded_method
        self.subscribers = []

    async def __call__(self, *args, **kwargs):
        await self.bounded_method(*args, **kwargs)
        tasks = [subscriber(self.bounded_method.__self__) for subscriber in self.subscribers]
        await asyncio.gather(*tasks)

    async def subscribe(self, callback, *, trigger: bool = True):
        self.subscribers.append(callback)
        if trigger:
            await callback(self.bounded_method.__self__)


class EffectorProtocol(Protocol):
    async def subscribe(self, callback, *, trigger: bool = True):
        ...

    async def __call__(self, *args, **kwargs):
        ...


def create_emitter(func) -> EffectorProtocol:
    _emitter_instance = Effector(func)

    async def _wrapper(*args, **kwargs):
        await _emitter_instance(*args, **kwargs)

    wrapper: EffectorProtocol = wraps(func)(_wrapper)
    wrapper.subscribe = _emitter_instance.subscribe
    return wrapper


def effector(func) -> EffectorProtocol:
    if not asyncio.iscoroutinefunction(func):
        raise ValueError("emitter can only be used on async functions")
    func.is_emitter = True
    return func


def inject_effectors(_object: object):
    if isinstance(_object, BaseModel) and not _object.model_config.get("extra", None) == "allow":
        raise ValueError("Pydantic models must be declared with extra='allow' to use inject_emitters")
    for attr in dir(_object):
        if attr not in ["__fields__", "__fields_set__", "__signature__"] and (
            attr not in _object.model_computed_fields.keys() if isinstance(_object, BaseModel) else True
        ):
            if callable(getattr(_object, attr)) and ismethod(getattr(_object, attr)):
                _method = getattr(_object, attr)
                if hasattr(_method, "is_emitter"):
                    wrapped_method = create_emitter(_method)
                    setattr(_object, attr, wrapped_method)



from typing import Awaitable, Callable

from pydantic import PrivateAttr

from schorle.elements.base.element import Element
from schorle.elements.tags import HTMLTag
from schorle.reactives.classes import Classes

OnClick = Callable[..., Awaitable]


class Button(Element):
    tag: HTMLTag = HTMLTag.BUTTON
    _base_classes: Classes = Classes("btn")
    _additional_callbacks: list = PrivateAttr(default_factory=list)
    classes: Classes = Classes()

    def add_callback(self, trigger: str, callback: Callable):
        self._additional_callbacks.append((trigger, callback))

    def get_triggers_and_methods(self):
        for trigger, method in super().get_triggers_and_methods():
            yield trigger, method
        for trigger, callback in self._additional_callbacks:
            yield trigger, callback



from pydantic import Field


def Attribute(*args, **kwargs):  # noqa: N802
    """
    This is a helper function to make it easier to annotate attributes on elements.
    :param args: same as pydantic.Field
    :param kwargs: same as pydantic.Field
    """
    if "alias" in kwargs:
        attribute_name = kwargs.pop("alias")
    else:
        attribute_name = None

    kwargs["json_schema_extra"] = {"attribute_name": attribute_name, "attribute": True}
    return Field(*args, **kwargs)



from schorle.elements.attribute import Attribute
from schorle.elements.base.element import Element
from schorle.elements.tags import HTMLTag


class Img(Element):
    tag: HTMLTag = HTMLTag.IMG
    src: str = Attribute(..., alias="src")
    alt: str = Attribute(..., alias="alt")



from enum import Enum


class HTMLTag(str, Enum):
    FIGURE = "figure"
    DEV_FOOTER = "dev-footer"
    FOOTER = "footer"
    HTML = "html"
    HEAD = "head"
    BODY = "body"
    P = "p"
    H1 = "h1"
    H2 = "h2"
    H3 = "h3"
    H4 = "h4"
    H5 = "h5"
    H6 = "h6"
    DIV = "div"
    SPAN = "span"
    A = "a"
    IMG = "img"
    UL = "ul"
    OL = "ol"
    LI = "li"
    TABLE = "table"
    TR = "tr"
    TD = "td"
    TH = "th"
    FORM = "form"
    INPUT = "input"
    SELECT = "select"
    OPTION = "option"
    BUTTON = "button"
    SCRIPT = "script"
    STYLE = "style"
    META = "meta"
    LINK = "link"
    TITLE = "title"



from typing import Optional

from pydantic import Field

from schorle.elements.base.base import BaseElement
from schorle.elements.base.element import Element
from schorle.elements.tags import HTMLTag


class Page(Element):
    tag: HTMLTag = HTMLTag.DIV
    element_id: str = "schorle-page"

    def find_by_id(self, element_id: str) -> Optional[BaseElement]:
        for child in self.traverse():
            if child.element_id == element_id:
                return child
        return None

    async def execute_all_before_render(self):
        self.inject_page_reference(self)

        for element in self.traverse():
            if isinstance(element, Element):
                await element.before_render()
                # cleanup render queues
                for attr in element.get_reactive_attributes():
                    while not attr._render_queue.empty():
                        await attr._render_queue.get()

        self.inject_page_reference(self)


def PageReference():  # noqa: N802
    return Field(None, json_schema_extra={"attribute": False, "page_reference": True})



from __future__ import annotations

from inspect import isclass
from types import GenericAlias, UnionType
from typing import Any, Iterator, get_origin

from loguru import logger
from lxml.etree import Element as LxmlElementFactory
from lxml.etree import _Element as LxmlElement
from lxml.etree import tostring
from pydantic import ConfigDict, Field, PrivateAttr
from pydantic.fields import FieldInfo

from schorle.elements.base.mixins import AttrsMixin, FactoryMixin
from schorle.elements.tags import HTMLTag
from schorle.reactives.base import ReactiveBase
from schorle.reactives.classes import Classes
from schorle.reactives.text import Text


class BaseElement(AttrsMixin, FactoryMixin):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")
    tag: HTMLTag
    text: Text | str = Field(default=Text(), description="Text content of the element, if any")
    style: dict[str, str] | None = Field(default=None, description="Style attributes of the element, if any")
    element_id: str | None = Field(default=None, description="Explicitly set the id of the element, if required")
    _rendering_element: LxmlElement | None = PrivateAttr(default=None)
    render_behaviour: str = Field(default="default", description="Render behaviour of the element")

    def _union_related_to_element(self, anno: UnionType):
        for arg in anno.__args__:
            if isclass(arg) and get_origin(arg) not in [list, dict] and issubclass(arg, BaseElement):
                return True
            elif isinstance(arg, UnionType):
                return self._union_related_to_element(arg)

    def _list_or_dict_related_to_element(self, anno: GenericAlias):
        for arg in anno.__args__:
            if isclass(arg) and issubclass(arg, BaseElement):
                return True
            elif isinstance(arg, UnionType):
                return self._union_related_to_element(arg)

    def _related_to_element(self, field: FieldInfo) -> bool:
        anno: type[Any] | None = field.annotation
        if get_origin(anno) in [list, dict] and type(anno) is GenericAlias:
            return self._list_or_dict_related_to_element(anno)  # type: ignore[arg-type]
        if isclass(anno) and issubclass(anno, BaseElement):
            return True
        elif isinstance(anno, UnionType) and self._union_related_to_element(anno):
            return True
        elif isclass(anno) and issubclass(anno, ReactiveBase) and not issubclass(anno, (Text, Classes)):
            return True
        else:
            return False

    def walk(self, parent: BaseElement | None = None) -> Iterator[tuple[BaseElement | None, BaseElement]]:
        """
        Traverse the element tree and yield a tuple of [parent, child] elements.
        """
        yield parent, self
        for k, v in self.model_fields.items():
            if v.json_schema_extra and v.json_schema_extra.get("page_reference"):
                continue
            elif self._related_to_element(v):
                element = getattr(self, k)
                if isinstance(element, BaseElement):
                    yield from element.walk(self)
                elif isinstance(element, ReactiveBase):
                    value = element.get()
                    if isinstance(value, list):
                        for _element in value:
                            yield from _element.walk(self)
                    elif isinstance(value, BaseElement):
                        yield from value.walk(self)
                elif isinstance(element, list):
                    for _element in element:
                        yield from _element.walk(self)

    def traverse(self, *, skip_self: bool = False) -> Iterator[BaseElement]:
        """
        Traverse the element tree and yield each element.
        """
        if not skip_self:
            yield self
        for k, v in self.model_fields.items():
            if v.json_schema_extra and v.json_schema_extra.get("page_reference"):
                continue
            elif self._related_to_element(v):
                element = getattr(self, k)
                if isinstance(element, BaseElement):
                    yield from element.traverse()
                elif isinstance(element, ReactiveBase):
                    value = element.get()
                    if isinstance(value, list):
                        for _element in value:
                            yield from _element.traverse()
                    elif isinstance(value, BaseElement):
                        yield from value.traverse()
                elif isinstance(element, list):
                    for _element in element:
                        yield from _element.traverse()

    def _add_classes(self, element: LxmlElement):
        pass

    def get_prerender(self) -> LxmlElement:
        if self._rendering_element is None:
            element = LxmlElementFactory(self.tag.value)

            if self.element_id is not None:
                element.set("id", self.element_id)

            if self.style is not None:
                element.set("style", ";".join([f"{k}:{v}" for k, v in self.style.items()]))

            self._add_classes(element)

            for k, v in self.get_element_attributes().items():
                if v is not None:
                    element.set(k, v)
            if self.text is not None:
                element.text = self.text.get() if isinstance(self.text, Text) else self.text
            self._rendering_element = element

        return self._rendering_element

    def render(self) -> str:
        root_element = None
        for parent, child in self.walk():
            if not parent:
                root_element = child.get_prerender()
            elif parent.render_behaviour == "default" and child.render_behaviour == "default":
                # default behaviour is to append the child to the parent
                parent.get_prerender().append(child.get_prerender())
            elif parent.render_behaviour == "flatten" and child.render_behaviour == "default":
                # flatten behaviour is to append the children of the child to the parent
                for sub_children in child.traverse(skip_self=True):
                    parent.get_prerender().append(sub_children.get_prerender())
            elif parent.render_behaviour == "default" and child.render_behaviour == "flatten":
                # flatten behaviour is to append the children of the child to the parent
                for sub_children in child.traverse(skip_self=True):
                    parent.get_prerender().append(sub_children.get_prerender())
            elif parent.render_behaviour == "flatten" and child.render_behaviour == "flatten":
                # in this case we need to get the parent of the parent and append the children of the child to it
                # this is because we need to flatten the children of the child to the parent of the parent
                parent_of_parent = parent.get_prerender().getparent()
                if parent_of_parent is not None:
                    for sub_children in child.traverse(skip_self=True):
                        parent_of_parent.append(sub_children.get_prerender())
                else:
                    logger.warning("Failed to get parent of parent")

        if root_element is None:
            msg = f"Failed to render element {self}."
            raise RuntimeError(msg)

        result = tostring(root_element, pretty_print=True).decode("utf-8")
        for element in self.traverse():
            element._rendering_element = None
        return result

    def __repr__(self):
        return f"<{self.tag} {self.element_id}>"

    def __str__(self):
        return self.__repr__()



import asyncio
import pkgutil
from asyncio import iscoroutinefunction
from functools import partial
from importlib.resources import files
from pathlib import Path
from typing import Callable, Union

from fastapi import FastAPI
from loguru import logger
from starlette.endpoints import WebSocketEndpoint
from starlette.responses import FileResponse, HTMLResponse, PlainTextResponse
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket

from schorle.elements.base.element import Element
from schorle.elements.html import BodyWithPage, EventHandler, Html, Meta, MorphWrapper
from schorle.elements.page import Page
from schorle.emitters import PageEmitter
from schorle.models import HtmxMessage
from schorle.theme import Theme
from schorle.utils import RunningMode, get_running_mode


def favicon() -> Union[FileResponse, HTMLResponse]:
    package_path = files("schorle")
    favicon_path = Path(str(package_path)) / "assets" / "favicon.svg"
    return FileResponse(favicon_path, media_type="image/svg+xml")


def assets(file_name: str) -> PlainTextResponse:
    _bundle = pkgutil.get_data("schorle", f"assets/{file_name}")
    if _bundle:
        return PlainTextResponse(_bundle.decode("utf-8"), status_code=200)
    else:
        return PlainTextResponse(f"File not found: {file_name}", status_code=404)


class Schorle:
    def __init__(self, theme: Theme = Theme.DARK) -> None:
        self._pages: dict[str, Page] = {}
        self.backend = FastAPI()
        self.backend.get("/_schorle/assets/{file_name:path}")(assets)
        self.backend.add_websocket_route("/_schorle/events", partial(EventsEndpoint, pages=self._pages))
        self.backend.get("/favicon.svg", response_model=None)(favicon)
        self.theme: Theme = theme

    def get(self, path: str):
        def decorator(func: Callable[..., Page]):
            self.backend.get(path, response_class=HTMLResponse)(partial(self.render_to_response, func))
            return func

        return decorator

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """
        This method is called by uvicorn when the server is started.
        """
        await self.backend(scope=scope, receive=receive, send=send)

    async def render_to_response(self, page_provider: Callable[..., Page]) -> HTMLResponse:
        page = page_provider()
        await page.execute_all_before_render()
        logger.info(f"Rendering page: {page}...")

        handler = EventHandler(content=page)
        body = BodyWithPage(wrapper=MorphWrapper(handler=handler))
        logger.debug(f"Rendering page: {page} with theme: {self.theme}...")
        html = Html(body=body, theme=self.theme)

        if get_running_mode() == RunningMode.DEV:
            logger.info("Adding dev meta tags...")
            html.head.dev_meta = Meta(name="schorle-dev", content="true")

        response = HTMLResponse(html.render(), status_code=200)
        logger.info(f"Adding page to cache with token: {html.head.csrf_meta.content}")

        self._pages[html.head.csrf_meta.content] = page

        logger.debug("Page rendered.")

        return response


class EventsEndpoint(WebSocketEndpoint):
    encoding = "text"

    def __init__(self, scope, receive, send, pages: dict[str, Page]) -> None:
        super().__init__(scope, receive, send)
        self._page: Page | None = None
        self._pages: dict[str, Page] = pages
        self.page_emitter_task = None

    async def on_connect(self, websocket: WebSocket) -> None:
        token = websocket.query_params.get("token")
        if not token:
            logger.error("No token provided.")
            await websocket.close()
            return
        page: Page | None = self._pages.get(token)
        if page:
            await websocket.accept()
            self._page = page
            self.page_emitter_task = asyncio.create_task(PageEmitter(page).emit(websocket))
            logger.info("Events connected.")

        elif not page and get_running_mode() == RunningMode.DEV:
            logger.info("Sending reload message to client...")
            await websocket.accept()
            await websocket.send_text("reload")  # standard schorle reload message
            await websocket.close()
            return
        else:
            logger.error(f"No page found for token: {token}")
            await websocket.close()
            return

    async def on_receive(self, ws: WebSocket, data: str) -> None:
        logger.warning(f"Events received message: {data}")
        message = HtmxMessage.model_validate_json(data)
        logger.debug(f"Events received message: {message}")
        if self._page:
            _element = self._page.find_by_id(message.headers.trigger_element_id)
            if _element and isinstance(_element, Element):
                logger.debug(f"Events found element: {_element}")
                method = _element.reactive_methods.get(message.headers.trigger_type)
                if not method:
                    msg = (
                        f"No method found for trigger: {message.headers.trigger_element_id} "
                        f"{message.headers.trigger_type}"
                    )
                    logger.warning(msg)
                else:
                    logger.debug(f"Events found method: {method}")
                    wrapped_method = wrap_in_coroutine(method)
                    if message.headers.trigger_type == "change" and _element.name:
                        new_value = getattr(message, _element.name)
                        await wrapped_method(new_value)
                    else:
                        await wrapped_method()
                    logger.debug(f"Events executed method: {method}")
            else:
                logger.error(f"No reactive element found for id: {message.headers.trigger_element_id}")
        else:
            logger.error("No page found, closing websocket...")
            await ws.close()

    async def on_disconnect(self, _: WebSocket, close_code: int) -> None:
        logger.info(f"Events disconnected with code: {close_code}")
        if self.page_emitter_task:
            self.page_emitter_task.cancel()
            logger.info("Events emitter task cancelled.")


def wrap_in_coroutine(func: Callable) -> Callable:
    if iscoroutinefunction(func):
        return func
    else:

        async def _wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return _wrapper
